function [A, C, Q, R0, R1, B, initx, initV, indicator,confidence, xsmooth] = ...
    learn_kalman(data, ct, A, C, Q, R1, R0, B, initx, initV, max_iter, affil, lambda1, lambda2, mapping, cutoff )

% input:
% data: Each row is a gene, each column is a time point. Multiple datasets are combined as data = [ds1 ds2 ds3 ...];
% ct: Cumulative timepoints. E.g., the first dataset has 8 time points, the second has 5 time points, then ct = [0 8 13]. I.e., data(:, ct(n)+1:ct(n+1)) is the n'th dataset.
% A, C, Q, R1, R0, B, initx, initV: The initial values for T_t = A * T_t-1, G_t = C * T_t. 
%                                   Q is the diagnoal matrix for TF activity smoothness, 
%                                   R0 and R1 are expression noise parameters for no imcoming edge and otherwise, respectively,
%                                   initix and initV are the mean and variance for the first timepoint.
% let ss be the # of TFs, os be the # of genes, nds be the # of datasets:
% A is ss by ss matrix
% C is os by ss matrix
% Q is ss by ss by nds matrix
% R1 and R0 are nds dim vector
% initx is ss by nds matrix
% initV is ss by ss by nds matrix

% max_iter : max number of outer iterations
% affil : Evidence association matrix
% lambda1 and lambda2: tradeoff constants
% mapping: E.g., mapping(i) = j,    i.e. data(j,:) is the gene expression for TF i;
% cutoff: adjustment for the likelihood comparison in PTM indicator prediction

% output:
% A, C, Q, R0, R1, B, initx, initV: learned values.
% indicator: inferred PTM indicator.
% confidence: confidence scores for each indicator.
% xsmooth: inferred activity levels for TFs.

% see paper for more details.

diagQ = 1;
diagR = 1;

if length(find(data > 100)) > 0
    fprintf('need to impute gene expression levels');
end

fprintf('%d genes selected for inferring the underlying processes.\n' , size(data,1));

maxTF = 3;
N = length(ct)-1;
help = 10 * ones(N,1);
verbose = 1;
option =  optimset('Display','off','LargeScale','off');

[os ss] = size(C);


alpha = zeros(os, os, N);
Tsum = zeros(N,1);
for d = 1:N
  %y = data(:,:,ex);
  y = data(:,ct(d)+1:ct(d+1));
  T = size(y,2);
  Tsum(d) = T;
  alpha_temp = zeros(os, os);
  for t=1:T
    alpha_temp = alpha_temp + y(:,t)*y(:,t)';
  end
  alpha(:,:,d) = alpha(:,:,d) + alpha_temp;
end

previous_loglik = -inf;
loglik = 0;
converged = 0;
num_iter = 1;
LL = [];


numTF = size(affil,2);
timept = size(data,2);
max_Estep = 5;
%%%   first initialize indicator

indicator = zeros(numTF,1);
indicator_prior = 0.8;   % to be post transcriptionally modified.

edge_prior = 0.7;

TFactivity = ones(numTF,timept) * 100;
ARmode = 0;
noedge = find(sum(C,2) == 0);

% pre-processing, kalman filter to infer all hidden value, use the most
% unlike curves as the post-transcriptional modification.
tempxsmooth = [];
tempmRNA = [];
for d = 1:N
    y = data(:,ct(d)+1:ct(d+1));
    Qt = Q(:,:,d);
    Vt = initV(:,:,d);
    Xt = initx(:,d);
    tempR = repmat(R1(d),os,1);
    tempR(noedge) = R0(d);
    tempR = diag(tempR);
    [beta_t, gamma_t, delta_t, gamma1_t, gamma2_t, x1, V1, loglik_t, xsmooth(:,ct(d)+1:ct(d+1)),help(d)] = ...
            Estep(y, eye(numTF), C , Qt, tempR, Xt, Vt, ARmode, num_iter,help(d));
    tempxsmooth = [tempxsmooth xsmooth(:,ct(d)+2:ct(d+1))];
    tempmRNA = [tempmRNA data(mapping,ct(d)+1:(ct(d+1)-1))];
end

for tf = 1:length(mapping)
    correlation(tf,1) = abs(corr(tempxsmooth(tf,:)' , tempmRNA(tf,:)'));
    correlation(tf,2) = tf;
end

result = sortrows(correlation,1);

indicator( result(1:floor(length(mapping)/5),2) ) = 1;

indicator_old_old = zeros(numTF,1);
xsmooth = zeros(ss,ct(length(ct)));
while sum(indicator_old_old ~= indicator)>0 && (num_iter <= max_iter) 

    num_iter =  num_iter + 1
    indicator_old_old = indicator;
  %%% E step
  
   
    numEstep = 1;
    indicator_old = ones(length(mapping),1);
  
    while numEstep <= max_Estep && sum(indicator_old ~= indicator)>0
        indicator'
        confidence = zeros(numTF,1);
        
        indicator_old = indicator;
        fprintf('in Estep: interation %d\n' , numEstep);
        numEstep = numEstep + 1;
        
        %%%   E.1  infer hidden chain based on indicator
        delta = zeros(os, ss,N);
        gamma = zeros(ss, ss,N);
        gamma1 = zeros(ss, ss,N);
        gamma2 = zeros(ss, ss,N);
        beta = zeros(ss, ss,N);
        P1sum = zeros(ss, ss,N);
        x1sum = zeros(ss, 1,N);
        loglik = 0;

        for TF = 1:length(mapping)
            tempC = C;
            tempC(:,TF) = 0;
            SubstractedData = data - tempC * xsmooth ;
            
            associatedgene = find(C(:,TF) > 0);
            indicator_posterior1 = 0;
            indicator_posterior0 = 0;
            nopostactivity = [];
            if length(associatedgene) ~= 0
                
                for d = 1:N
                    y = SubstractedData(associatedgene,ct(d)+1:ct(d+1));
                    Qt = Q(TF,TF,d);
                    Vt = initV(TF,TF,d);
                    Xt = initx(TF,d);
                    tempR = repmat(R1(d),length(associatedgene),1);
                    tempR = diag(tempR);

                    [beta_t, gamma_t, delta_t, gamma1_t, gamma2_t, x1, V1, loglik_t, xsmooth(TF,ct(d)+1:ct(d+1)),help(d)] = ...
                    Estep(y, 1, C(associatedgene,TF) , Qt, tempR, Xt, Vt, ARmode, num_iter,help(d));

                    beta(TF,TF,d) = beta_t;
           %         gamma = gamma + gamma_t;
                    delta(associatedgene,TF,d) = delta_t;
                    gamma1(TF,TF,d) = gamma1_t;
                    gamma2(TF,TF,d) = gamma2_t;
                    P1sum(TF,TF,d) = V1 + x1*x1';
                    x1sum(TF,d) = x1;

                    
                    
                    % compute no post-transcriptional model
                    W = C(associatedgene,TF);
                    mu = data(mapping(TF), ct(d)+1:(ct(d+1)-1));

                    nopostmod_mu = mu + (W * B(d))' ...
                            * inv(W * B(d) * W' + diag(repmat(R1(d),length(associatedgene),1))) ...
                            * (SubstractedData(associatedgene, ct(d)+2:ct(d+1))  - W * mu);
                    
                    indicator_posterior0 = indicator_posterior0 - sum((nopostmod_mu - mu) .^ 2 / R1(d) / 2) ...
                                            - log(R1(d)) / 2 * (ct(d+1) - ct(d) - 1) ...
                                            - sum(sum((SubstractedData(associatedgene, ct(d)+2:ct(d+1))  - W * nopostmod_mu).^2)) / R1(d)...
                                            - length(associatedgene) * log(R1(d)) / 2 * (ct(d+1) - ct(d) - 1);
                                        
                    indicator_posterior1 = indicator_posterior1 ...
                                            - sum( (xsmooth(TF,(ct(d)+2):(ct(d+1))) - xsmooth(TF,(ct(d)+1):(ct(d+1)-1))).^2 , 2 ) / (2 * Qt)  ...
                                            - log(Qt) / 2 * (ct(d+1) - ct(d) - 1) ...
                                            - sum(sum((SubstractedData(associatedgene, ct(d)+2:ct(d+1))  - W * xsmooth(TF,(ct(d)+2):(ct(d+1)))).^2)) / R1(d)...
                                            - length(associatedgene) * log(R1(d)) / 2 * (ct(d+1) - ct(d) - 1);
                    nopostactivity = [nopostactivity 0 nopostmod_mu];

                end
                fprintf('difference = %g\n',indicator_posterior1 - indicator_posterior0);
                indicator_posterior1 = indicator_posterior1 - cutoff;                
                if indicator_posterior0 > indicator_posterior1
                    fprintf('0 > 1 TF = %d\n\n',TF);
                    indicator(TF) = 0;
                    xsmooth(TF,:) = nopostactivity;
                    confidence(TF) = indicator_posterior0 - indicator_posterior1;
                    
                else
                    indicator(TF) = 1;
                    confidence(TF) = indicator_posterior1 - indicator_posterior0;
                   
                end
                
            else        % no genes associated
                
                indicator(TF) = 0;
                
                
                for d = 1:N
                    xsmooth(TF,ct(d)+1) = 0;
                    xsmooth(TF,ct(d)+2:ct(d+1)) = data(mapping(TF) ,ct(d)+1:(ct(d+1)-1) );
                end
            end
                    

        end
        
        
    end     
        
        
        
  

  %%% M step
  
  % Tsum =  N*T
  % Tsum1 = N*(T-1);
  Tsum1 = sum(Tsum) - N;
  % %%% A = beta * inv(gamma1);
  %A = (gamma1' \ beta')';
%  for d = 1:N
%      Q(:,:,d) = (gamma2(:,:,d) - A*beta(:,:,d)') / Tsum(d);
%      if diagQ
%        Q(:,:,d) = diag(repmat(mean(diag(Q(:,:,d))),length(Q(:,:,d)),1));
%      end
%  end

    noedge = [];
       fprintf('calculate gene ');
    for i = 1:os


        value = zeros(maxTF,1);

        value0 = 0;
        weight = [];
        for d = 1: N
            value0 = value0 + mean( (data(i,ct(d)+1:ct(d+1)).^2)' / R0(d)) / N   ;
            weight = [weight; ones(ct(d+1) - ct(d),1) /  R0(d) ];
        end

        value0 = value0 + lambda1 * sum((1-affil(i,:)) * (1 - edge_prior) + affil(i,:) * edge_prior);
        currentTF = [];
        localweight = zeros(maxTF,ss,maxTF);

        for assocTF = 1:maxTF
            for p = 1:ss
                if sum(p == currentTF) > 0
                    eval(p) = Inf;
                    continue;
                end

                cur = [currentTF p];  
                X = xsmooth(cur,:)';
                Y = data(i,:)';
                localweight(1:assocTF,p,assocTF) = inv(X' * diag(weight) * X) * X' * diag(weight) * Y;
                eval(p) = mean(((Y -  X * localweight(1:assocTF,p,assocTF) ).^2) .* weight);

                proc = cur;
                tempaffil = (1 - edge_prior) * ones(length(affil(i,:)),1);
                tempaffil(proc(find(affil(i,proc) == 0))) = edge_prior;
                proc_not_inuse = setdiff(1:length(affil(i,:)) , proc);
                tempaffil(proc_not_inuse(find(affil(i,proc_not_inuse) == 1))) = edge_prior;
                eval(p) = eval(p) + lambda1 * sum(tempaffil) + lambda2 * sum(abs(localweight(1:assocTF,p,assocTF)));
            end
            [value(assocTF) , index] = min(eval);
            currentTF = [currentTF index];
        end


        C(i,:) = 0;   

        [j , index] = min(value);
        
        if j < value0
            C(i,currentTF(1:index)) = localweight(1:index,currentTF(index),index);
        else
            noedge = [noedge i];
        end
        

            fprintf('%d ',i);
            if rem(i,30) == 0
                fprintf('\n');
            end


    end
       fprintf('\n');






    tempR = zeros(os,os,N);
    for d = 1:N
        initx(:,d) = mean(x1sum(:,d));
        initV(:,:,d) = P1sum(:,:,d) / 1 - x1sum(:,d)*x1sum(:,d)';
        tempV = diag(initV(:,:,d));

        initV(:,:,d) = diag( repmat( mean(tempV) , length(initV(:,:,d)) , 1 ) );
        tempR(:,:,d) = (alpha(:,:,d) - C*delta(:,:,d)') / Tsum(d);
        if diagR
            wholegene = 1:os;
            foredgedgene = setdiff(wholegene,noedge);
 %           R1(d) = mean(diag(tempR(foredgedgene,foredgedgene,d)));
            if length(noedge) ~=0
 %               R0(d) = mean(diag(tempR(noedge,noedge,d)));
            end
        end


 %       B(d) = R1(d);


    end





%   if ~isempty(constr_fun)
%     [A,C,Q,R,initx,initV] = feval(constr_fun, A, C, Q, R, initx, initV, varargin{:});
%   end
  
 % converged = em_converged(loglik, previous_loglik, thresh);
  previous_loglik = loglik;
end




%%%%%%%%%

function [beta, gamma, delta, gamma1, gamma2, x1, V1, loglik, xsmooth,help] = ...
    Estep(y, A, C, Q, R, initx, initV, ARmode , num_iter,help)
%
% Compute the (expected) sufficient statistics for a single Kalman filter sequence.
%

[os T] = size(y);
ss = length(A);

if ARmode
  xsmooth = y;
  Vsmooth = zeros(ss, ss, T); % no uncertainty about the hidden states
  VVsmooth = zeros(ss, ss, T);
  loglik = 0;
else
  [xsmooth, Vsmooth, VVsmooth, loglik,help] = kalman_smoother(y, A, C, Q, R, initx, initV, num_iter,help);
end

delta = zeros(os, ss);
gamma = zeros(ss, ss);
beta = zeros(ss, ss);
for t=1:T
  delta = delta + y(:,t)*xsmooth(:,t)';
  gamma = gamma + xsmooth(:,t)*xsmooth(:,t)' + Vsmooth(:,:,t);
  if t>1 beta = beta + xsmooth(:,t)*xsmooth(:,t-1)' + VVsmooth(:,:,t); end
end
gamma1 = gamma - xsmooth(:,T)*xsmooth(:,T)' - Vsmooth(:,:,T);
gamma2 = gamma - xsmooth(:,1)*xsmooth(:,1)' - Vsmooth(:,:,1);

x1 = xsmooth(:,1);
V1 = Vsmooth(:,:,1);

